/*
 *	Copyright (C) 2009, 2010. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file
@ingroup ExamplePrograms
@brief Reads image frames from a video source, and writes each one of them to an image file.

This is an example usage of the @ref QVMPlayerReader class, to perform image input/output from a non block-oriented application.

@section Usage
Simply compile, and execute the binary:
\code
	./imageIO <URL>
\endcode

@param URL Video source identificator. See section @ref QVMPlayerURLFormat for further info about its format.

@author PARP Research Group. University of Murcia, Spain.
*/
#include <QVMPlayerReader>
#include <qvip.h>

int main(int argc, char *argv[])
        {
	// 1. Check correct command line usage of the application.
	if (argc != 2)
		{
		std::cout << "Usage:" << std::endl << "\t" << argv[0] << " <video source url>" << std::endl;
		std::cout << std::endl << "Example:" << std::endl << "\t" << argv[0] << " penguin.avi" << std::endl;
		exit(0);
		}

	// 2. Opens the video source for reading.
	QVMPlayerReader videoReader;
	if (!videoReader.open(argv[1], QVMPlayerReader::NoLoop))
		{
		std::cout << "Error opening video source." << std::endl;
		exit(0);
		}
	std::cout << "Video source successfully opened." << std::endl;

	// 3. Reads each input image frame of the video source.
	int n = 0;
	QVImage<uChar, 3> image;
	while(videoReader.grab(image))
		{
		const QString imageFileName = QString("frame") + QString::number(n++).rightJustified(6, '0') + QString(".png");
		std::cout << "Frame grabbed ok. Writting frame to file " << qPrintable(imageFileName) << std::endl;
		writeQVImageToFile (imageFileName, image);
		}

	// 4. Finishes.
	videoReader.close();
	std::cout << "Video reading finished." << std::endl;
	}
